/*****************************************************************************

	Fox Counter for METAL GEAR SOLID PEACE WALKER

	Copyright (C) 2012 Yumimint

*****************************************************************************/
#include <windows.h>
#include <commctrl.h>
#ifdef _MSC_VER
#	pragma comment (lib, "comctl32.lib")
#endif

#include <tchar.h>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <cassert>
#include "resource.h"

#include <stack>

struct ACTION {
	ACTION() {}
	ACTION(int weapon_category, int count, bool kill)
		: weapon_category(weapon_category)
		, count(count)
		, kill(kill)
	{}

	int weapon_category;
	int count;
	bool kill;

	void display() const;

	bool issame(const ACTION &ref) const {
		return weapon_category == ref.weapon_category &&
			   kill == ref.kill;
	}
};

enum {
#	define X(a, b) IDX_##a,
#	include "table.h"
#	undef X
};

static const int CATEGORY_NUM(11);
static const int COUNTER_NUM(11+2);

static HWND   s_edit;
static HBRUSH s_hbrEdit1;
static HBRUSH s_hbrEdit2;
static int   s_addvalue = 1;
static int   s_maxnum;
static int   s_counter[COUNTER_NUM] = {0};
static float s_rate   [CATEGORY_NUM];
static std::stack<ACTION> s_action;
static std::stack<ACTION> s_redo_stack;
static std::stack<ACTION> s_disp_stack; // d폜\p
static std::stack<WPARAM> s_linepos;

static const char DATANAME[] = "FoxCounter.txt";

static const char *const catname[] = {
#	define X(a, b) b,
#	include "table.h"
#	undef X
};

static LRESULT CALLBACK dialog_proc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT command(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static void update(HWND hwnd);
static void load();
static void save();
static void log_clear();
static void log_remove_lastline();
static void log_vprintf(const char *fmt, va_list arglist);
static void log_printf(const char *fmt, ...);

template <typename T>
static void purge(T& x)
{
	while (!x.empty()) {
		x.pop();
	}
}


/*****************************************************************************
 *	main
 */
int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	InitCommonControls();
	DialogBox(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), NULL, (DLGPROC) dialog_proc);
	return 0;
}

/**
 */
static void apply(const ACTION &act)
{
	s_counter[act.kill ? IDX_KILL : IDX_NOKILL] += act.count;
	s_counter[act.weapon_category]              += act.count;

	if (!s_disp_stack.empty()) {
		ACTION &top(s_disp_stack.top());
		if (top.issame(act)) {
			top.count += act.count;
			log_remove_lastline();
			top.display();
			return;
		}
	}
	s_disp_stack.push(act);
	act.display();
}

/**
 */
static void revert(const ACTION &act)
{
	s_counter[act.kill ? IDX_KILL : IDX_NOKILL] -= act.count;
	s_counter[act.weapon_category]              -= act.count;

	log_remove_lastline();
	if (!s_disp_stack.empty()) {
		ACTION &top(s_disp_stack.top());
		if (top.issame(act)) {
			top.count -= act.count;
			if (top.count > 0) {
				top.display();
			} else {
				s_disp_stack.pop();
			}
		}
	}
}

/**
 */
void ACTION::display() const
{
	if (kill) {
		log_printf("%s EQ +%d\n", catname[weapon_category], count);
	} else {
		log_printf("%s ͉ +%d\n", catname[weapon_category], count);
	}
}


/**
 */
static LRESULT CALLBACK dialog_proc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
	case WM_INITDIALOG:
		s_edit = GetDlgItem(hwnd, IDC_EDIT_LOG);
		s_hbrEdit1 = CreateSolidBrush(RGB(128, 255, 128));
		s_hbrEdit2 = CreateSolidBrush(RGB(255, 128, 128));
		SendDlgItemMessage(hwnd, IDC_EDIT_VALUE, EM_SETLIMITTEXT, 6, 0);
		SendDlgItemMessage(hwnd, IDC_RADIO1, BM_CLICK, 0, 0);
		for (int i(0); i < CATEGORY_NUM; ++i) {
			SendDlgItemMessageA(hwnd, IDC_CAT1+i, WM_SETTEXT, 0, (LPARAM) catname[i]);
		}
		load();
		update(hwnd);
		return 1;

	case WM_DESTROY:
		DeleteObject(s_hbrEdit1);
		DeleteObject(s_hbrEdit2);
		return 0;

	case WM_CTLCOLORSTATIC:
		for (int i(0); i < CATEGORY_NUM; ++i) {
			if ((HWND)lParam == GetDlgItem(hwnd, IDC_EDIT_R1+i)) {
				if (s_rate[i] <= 1.1f && s_rate[i] >= 0.9f) {
					SetBkMode((HDC) wParam, TRANSPARENT);
					SetTextColor((HDC) wParam, RGB(0, 0, 0));
					return (LRESULT) s_hbrEdit1;
				}
			}
			if ((HWND)lParam == GetDlgItem(hwnd, IDC_EDIT_C1+i)) {
				if (s_counter[i] == s_maxnum) {
					SetBkMode((HDC) wParam, TRANSPARENT);
					SetTextColor((HDC) wParam, RGB(255, 255, 0));
					return (LRESULT) s_hbrEdit2;
				}
			}
		}
		break;

	case WM_COMMAND:
		return command(hwnd, msg, wParam, lParam);
	}
	return 0;
}

/**
 */
static LRESULT command(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (LOWORD(wParam)) {
	case IDCANCEL:
		EndDialog(hwnd, 0);
		break;

	case IDC_BUTTON_FIX:
		purge(s_action);
		purge(s_redo_stack);
		purge(s_disp_stack);
		update(hwnd);
		save();
		log_printf("---- m ----\n");
		break;

	case IDC_BUTTON_UNDO:
		if (!s_action.empty()) {
			revert(s_action.top());
			s_redo_stack.push(s_action.top());
			s_action.pop();
			update(hwnd);
		}
		break;

	case IDC_BUTTON_REDO:
		if (!s_redo_stack.empty()) {
			apply(s_redo_stack.top());
			s_action.push(s_redo_stack.top());
			s_redo_stack.pop();
			update(hwnd);
		}
		update(hwnd);
		break;

	case IDC_BUTTON_CLEAR:
		while (!s_action.empty()) {
			revert(s_action.top());
			s_redo_stack.push(s_action.top());
			s_action.pop();
		}
		update(hwnd);
		//log_clear();
		break;

	case IDC_RADIO1:
		s_addvalue = 1;
		EnableWindow(GetDlgItem(hwnd, IDC_EDIT_VALUE), false);
		break;

	case IDC_RADIO2:
		EnableWindow(GetDlgItem(hwnd, IDC_EDIT_VALUE), true);
		{
			char buf[32];
			GetWindowTextA(GetDlgItem(hwnd, IDC_EDIT_VALUE), buf, sizeof(buf));
			s_addvalue = atoi(buf);
		}
		break;

	case IDC_EDIT_VALUE:
		if (HIWORD(wParam) == EN_UPDATE) {
			char buf[32];
			GetWindowTextA(GetDlgItem(hwnd, IDC_EDIT_VALUE), buf, sizeof(buf));
			s_addvalue = atoi(buf);
		}
		break;

	default:
		if ((LOWORD(wParam) >= IDC_BUTTON_A1 && LOWORD(wParam) <= IDC_BUTTON_AB) ||
			(LOWORD(wParam) >= IDC_BUTTON_B1 && LOWORD(wParam) <= IDC_BUTTON_BB))
		{
			ACTION act;
			if ((LOWORD(wParam) >= IDC_BUTTON_A1 && LOWORD(wParam) <= IDC_BUTTON_AB)) {
				act = ACTION(LOWORD(wParam) - IDC_BUTTON_A1, s_addvalue, false);
			} else {
				act = ACTION(LOWORD(wParam) - IDC_BUTTON_B1, s_addvalue, true);
			}
			if (act.count > 0) {
				s_action.push(act);
				apply(act);
				purge(s_redo_stack);
				update(hwnd);
			}
		}
		break;
	}
	return 1;
}

/**
 */
static LRESULT settextf(HWND hDlg, int nIDDlgItem, const char *fmt, ...)
{
	LRESULT res;
	char buf[1024];
	va_list arglist; 
	va_start(arglist, fmt); 
	vsnprintf(buf, sizeof(buf), fmt, arglist);
	res = SendDlgItemMessageA(hDlg, nIDDlgItem, WM_SETTEXT, 0, (LPARAM) buf);
	va_end(arglist);
	return res;
}


/**
 *		\XV
 */
static void update(HWND hwnd)
{
	double sum(0);
	s_maxnum = 1;

	for (int i(0); i < CATEGORY_NUM; ++i) {
		if (s_maxnum < s_counter[i])
			s_maxnum = s_counter[i];
		sum += s_counter[i];
	}

	double avg(sum / CATEGORY_NUM);
	if (avg == 0.0) avg = 1;

	for (int i(0); i < CATEGORY_NUM; ++i) {
		const float r(s_counter[i] / avg);
		s_rate[i] = r;
		settextf(hwnd, IDC_EDIT_C1 + i, "%d"  , s_counter[i]);
		settextf(hwnd, IDC_EDIT_R1 + i, "%.4f", r);
	}

	settextf(hwnd, IDC_EDIT_NOKILL, "%d", s_counter[IDX_NOKILL]);
	settextf(hwnd, IDC_EDIT_KILL,   "%d", s_counter[IDX_KILL]*2);
	settextf(hwnd, IDC_EDIT_DIFF,   "%d", s_counter[IDX_NOKILL] - s_counter[IDX_KILL]*2);

	EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_REDO ), !s_redo_stack.empty());
	EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_UNDO ), !s_action.empty());
	EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_FIX  ), !s_action.empty());
	EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_CLEAR), !s_action.empty());
}

static int varindex(const char *name)
{
	for (int i(0); i < COUNTER_NUM; ++i) {
		if (strcmp(name, catname[i]) == 0)
			return i;
	}
	return -1;
}

/**
 */
static void load()
{
	FILE *fp = fopen(DATANAME, "rt");
	for (int i(0); i < COUNTER_NUM; ++i) {
		s_counter[i] = 0;
	}
	if (!fp) return;
	char buf[1024];
	while (fgets(buf, sizeof(buf), fp)) {
		int v(0);
		char varname[32];
		if (sscanf(buf, "%d %31s", &v, varname) == 2) {
			int i = varindex(varname);
//			log_printf("%s %d\n", varname, i);
			if (i >= 0) {
				s_counter[i] = v;
			}
		}
	}
	fclose(fp);
}

/**
 */
static void save()
{
	FILE *fp = fopen(DATANAME, "wt");
	if (!fp) {
		log_printf("%s: %s\n", DATANAME, strerror(errno));
		return;
	}
	for (int i(0); i < COUNTER_NUM; ++i) {
		fprintf(fp, "%d %s\n", s_counter[i], catname[i]);
	}
	fclose(fp);
}

/**
 */
static void log_clear()
{
	if (!s_edit) return;
	SendMessage(s_edit, EM_SETSEL, 0, -1);
	SendMessage(s_edit, EM_REPLACESEL, 0, 0);
	SendMessage(s_edit, WM_PAINT, 0, 0);
	purge(s_linepos);
}

/**
 */
static void log_remove_lastline()
{
	if (s_linepos.empty()) return;
	WPARAM lSize = SendMessage(s_edit, WM_GETTEXTLENGTH, 0, 0);
	SendMessage(s_edit, EM_SETSEL, s_linepos.top(), lSize);
	SendMessage(s_edit, EM_REPLACESEL, 0, 0);
	SendMessage(s_edit, WM_PAINT, 0, 0);
	s_linepos.pop();
}

/**
 */
static void log_vprintf(const char *fmt, va_list arglist)
{
	if (!s_edit) return;
	char buf[1024];
	vsnprintf(buf, sizeof(buf), fmt, arglist);
	WPARAM lSize = SendMessage(s_edit, WM_GETTEXTLENGTH, 0, 0);
	SendMessage(s_edit, EM_SETSEL, lSize, lSize);
	s_linepos.push(lSize);
	SendMessageA(s_edit, EM_REPLACESEL, 0, (LPARAM) buf);
}

/**
 */
static void log_printf(const char *fmt, ...)
{
	va_list arglist; 
	va_start(arglist, fmt); 
	log_vprintf(fmt, arglist);
	va_end(arglist); 
}

